from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from alaric.abc import ComparisonT


class CI:
    """
    Marks a given option as case-insensitive.
    """

    pass
